Jacchia-Bowman 2008
===================

```@meta
CurrentModule = SatelliteToolboxAtmosphericModels
DocTestSetup = quote
    using SatelliteToolboxAtmosphericModels
end
```

This is an empirical thermospheric density model based on the Jacchia theory. It was
published in:

> **Bowman, B. R., Tobiska, W. K., Marcos, F. A., Huang, C. Y., Lin, C. S., Burke, W. J
> (2008)**. *A new empirical thermospheric density model JB2008 using new solar and
> geomagnetic indices.* **In the proeceeding of the AIAA/AAS Astrodynamics Specialist
> Conference**, Honolulu, Hawaii.

For more information, visit
[http://sol.spacenvironment.net/jb2008](http://sol.spacenvironment.net/jb2008).

In this package, we can evaluate the model using the following functions:

```julia
function AtmosphericModels.jb2008(instant::DateTime, ϕ_gd::Number, λ::Number, h::Number[, F10::Number, F10ₐ::Number, S10::Number, S10ₐ::Number, M10::Number, M10ₐ::Number, Y10::Number, Y10ₐ::Number, DstΔTc::Number]) -> JB2008Output{Float64}
function AtmosphericModels.jb2008(jd::Number, ϕ_gd::Number, λ::Number, h::Number[, F10::Number, F10ₐ::Number, S10::Number, S10ₐ::Number, M10::Number, M10ₐ::Number, Y10::Number, Y10ₐ::Number, DstΔTc::Number]) -> JB2008Output{Float64}
```

where:

- `jd::Number`: Julian day to compute the model.
- `instant::DateTime`: Instant to compute the model represent using `DateTime`.
- `ϕ_gd`: Geodetic latitude [rad].
- `λ`: Longitude [rad].
- `h`: Altitude [m].
- `F10`: 10.7-cm solar flux [sfu] obtained 1 day before `jd`.
- `F10ₐ`: 10.7-cm averaged solar flux using a 81-day window centered on input time obtained
    1 day before `jd`.
- `S10`: EUV index (26-34 nm) scaled to F10.7 obtained 1 day before `jd`.
- `S10ₐ`: EUV 81-day averaged centered index obtained 1 day before `jd`.
- `M10`: MG2 index scaled to F10.7 obtained 2 days before `jd`.
- `M10ₐ`: MG2 81-day averaged centered index obtained 2 day before `jd`.
- `Y10`: Solar X-ray & Ly-α index scaled to F10.7 obtained 5 days before `jd`.
- `Y10ₐ`: Solar X-ray & Ly-α 81-day averaged centered index obtained 5 days before `jd`.
- `DstΔTc`: Temperature variation related to the Dst.

If we omit all space indices, the system tries to obtain them automatically for the selected
day `jd` or `instant`. However, the indices must be already initialized using the function
`SpaceIndices.init()`.

These functions return an object of type `JB2008Output{Float64}` that contains the
following fields:

- `total_density::T`: Total atmospheric density [1 / m³].
- `temperature::T`: Temperature at the selected position [K].
- `exospheric_temperature::T`: Exospheric temperature [K].
- `N2_number_density::T`: Number density of N₂ [1 / m³].
- `O2_number_density::T`: Number density of O₂ [1 / m³].
- `O_number_density::T`: Number density of O [1 / m³].
- `Ar_number_density::T`: Number density of Ar [1 / m³].
- `He_number_density::T`: Number density of He [1 / m³].
- `H_number_density::T`: Number density of H [1 / m³].

## Examples

```jldoctest
julia> AtmosphericModels.jb2008(DateTime("2022-06-19T18:35:00"), deg2rad(-22), deg2rad(-45), 700e3, 79, 73.5, 55.1, 53.8, 78.9, 73.3, 80.2, 71.7, 50)
Jacchia-Bowman 2008 Atmospheric Model Result:
      Total density :    5.19343e-15  kg / m³
        Temperature :         826.77  K
   Exospheric Temp. :         819.72  K
  N₂ number density :    2.65437e+07  1 / m³
  O₂ number density :         193998  1 / m³
  O  number density :    7.67489e+10  1 / m³
  Ar number density :        13.3774  1 / m³
  He number density :    4.64209e+11  1 / m³
  H  number density :    4.07234e+10  1 / m³
```

```julia-repl
julia> SpaceIndices.init()

julia> AtmosphericModels.jb2008(DateTime("2022-06-19T18:35:00"), deg2rad(-22), deg2rad(-45), 700e3)
Jacchia-Bowman 2008 Atmospheric Model Result:
      Total density :    1.98151e-14  kg / m³
        Temperature :        1092.64  K
   Exospheric Temp. :        1082.65  K
  N₂ number density :    1.73067e+09  1 / m³
  O₂ number density :    2.53444e+07  1 / m³
  O  number density :    6.15818e+11  1 / m³
  Ar number density :        6990.11  1 / m³
  He number density :    5.06517e+11  1 / m³
  H  number density :    4.32842e+09  1 / m³
```

If we use the automatic space index fetching mechanism, it is possible to obtain the fetched
values by turning on the debugging logs according to the [Julia
documentation](https://docs.julialang.org/en/v1/stdlib/Logging/):

```julia-repl
julia> using Logging

julia> with_logger(ConsoleLogger(stderr, Logging.Debug)) do
           AtmosphericModels.jb2008(DateTime("2022-06-19T18:35:00"), deg2rad(-22), deg2rad(-45), 700e3)
       end
┌ Debug: JB2008 - Fetched Space Indices
│   Daily F10.7           : 140.2 sfu
│   81-day averaged F10.7 : 127.25679012345678 sfu
│   Daily S10             : 107.1
│   81-day averaged S10   : 98.8
│   Daily M10             : 165.6
│   81-day averaged M10   : 139.6
│   Daily Y10             : 150.7
│   81-day averaged Y10   : 143.7
│   Exo. temp. variation  : 64.58333333333333
└ @ SatelliteToolboxAtmosphericModels.AtmosphericModels ~/tmp/SatelliteToolboxAtmosphericModels.jl/src/jb2008/jb2008.jl:90
Jacchia-Bowman 2008 Atmospheric Model Result:
      Total density :    1.98151e-14  kg / m³
        Temperature :        1092.64  K
   Exospheric Temp. :        1082.65  K
  N₂ number density :    1.73067e+09  1 / m³
  O₂ number density :    2.53444e+07  1 / m³
  O  number density :    6.15818e+11  1 / m³
  Ar number density :        6990.11  1 / m³
  He number density :    5.06517e+11  1 / m³
  H  number density :    4.32842e+09  1 / m³
```
