Jacchia-Roberts 1971
====================

```@meta
CurrentModule = SatelliteToolboxAtmosphericModels
DocTestSetup = quote
    using SatelliteToolboxAtmosphericModels
end
```

This is an analytic atmospheric model based on the Jacchia's 1970 model. It was published
in:

> **Roberts, C. E (1971)**. *An analytic model for upper atmosphere densities based upon
> Jacchia's 1970 models*. **Celestial mechanics**, Vol. 4 (3-4), p. 368-377, DOI:
> 10.1007/BF01231398.

Although it is old, this model is still used for some applications, like computing the
estimated reentry time for an object on low Earth orbits.

In this package, we can evaluate the model using the following functions:

```julia
function AtmosphericModels.jr1971(instant::DateTime, ϕ_gd::Number, λ::Number, h::Number[, F10::Number, F10ₐ::Number, Kp::Number])
function AtmosphericModels.jr1971(jd::Number, ϕ_gd::Number, λ::Number, h::Number[, F10::Number, F10ₐ::Number, Kp::Number])
```

where:

- `jd::Number`: Julian day to compute the model.
- `instant::DateTime`: Instant to compute the model represent using `DateTime`.
- `ϕ_gd::Number`: Geodetic latitude [rad].
- `λ::Number`: Longitude [rad].
- `h::Number`: Altitude [m].
- `F10::Number`: 10.7-cm solar flux [sfu].
- `F10ₐ::Number`: 10.7-cm averaged solar flux, 81-day centered on input time [sfu].
- `Kp::Number`: Kp geomagnetic index with a delay of 3 hours.

If we omit all space indices, the system tries to obtain them automatically for the selected
day `jd` or `instant`. However, the indices must be already initialized using the function
`SpaceIndices.init()`.

This function return an object of type `JR1971Output{Float64}` that contains the following
fields:

- `total_density::T`: Total atmospheric density [1 / m³].
- `temperature::T`: Temperature at the selected position [K].
- `exospheric_temperature::T`: Exospheric temperature [K].
- `N2_number_density::T`: Number density of N₂ [1 / m³].
- `O2_number_density::T`: Number density of O₂ [1 / m³].
- `O_number_density::T`: Number density of O [1 / m³].
- `Ar_number_density::T`: Number density of Ar [1 / m³].
- `He_number_density::T`: Number density of He [1 / m³].
- `H_number_density::T`: Number density of H [1 / m³].

## Examples

```jldoctest
julia> AtmosphericModels.jr1971(DateTime("2018-06-19T18:35:00"), deg2rad(-22), deg2rad(-45), 700e3, 79, 73.5, 1.34)
Jacchia-Roberts 1971 Atmospheric Model Result:
      Total density :     9.6849e-15  kg / m³
        Temperature :         832.02  K
   Exospheric Temp. :         832.02  K
  N₂ number density :     2.8435e+07  1 / m³
  O₂ number density :         174223  1 / m³
  O  number density :    1.41391e+11  1 / m³
  Ar number density :        8.97257  1 / m³
  He number density :    8.77389e+11  1 / m³
  H  number density :    5.70278e+10  1 / m³
```

```julia-repl
julia> SpaceIndices.init()

julia> AtmosphericModels.jr1971(DateTime("2022-06-19T18:35:00"), deg2rad(-22), deg2rad(-45), 700e3)
Jacchia-Roberts 1971 Atmospheric Model Result:
      Total density :    6.03113e-14  kg / m³
        Temperature :        1108.65  K
   Exospheric Temp. :        1108.69  K
  N₂ number density :    3.28087e+09  1 / m³
  O₂ number density :    4.11164e+07  1 / m³
  O  number density :    1.88625e+12  1 / m³
  Ar number density :        8826.28  1 / m³
  He number density :    1.50861e+12  1 / m³
  H  number density :    9.14578e+09  1 / m³
```

If we use the automatic space index fetching mechanism, it is possible to obtain the fetched
values by turning on the debugging logs according to the [Julia
documentation](https://docs.julialang.org/en/v1/stdlib/Logging/):

```julia-repl
julia> with_logger(ConsoleLogger(stderr, Logging.Debug)) do
           AtmosphericModels.jr1971(DateTime("2022-06-19T18:35:00"), deg2rad(-22), deg2rad(-45), 700e3)
       end
┌ Debug: JR1971 - Fetched Space Indices
│   Daily F10.7           : 143.6 sfu
│   81-day averaged F10.7 : 126.9604938271605 sfu
│   3-hour delayed Kp     : 1.333
└ @ SatelliteToolboxAtmosphericModels.AtmosphericModels ~/tmp/SatelliteToolboxAtmosphericModels.jl/src/jr1971/jr1971.jl:81
Jacchia-Roberts 1971 Atmospheric Model Result:
      Total density :    6.03113e-14  kg / m³
        Temperature :        1108.65  K
   Exospheric Temp. :        1108.69  K
  N₂ number density :    3.28087e+09  1 / m³
  O₂ number density :    4.11164e+07  1 / m³
  O  number density :    1.88625e+12  1 / m³
  Ar number density :        8826.28  1 / m³
  He number density :    1.50861e+12  1 / m³
  H  number density :    9.14578e+09  1 / m³
```
