# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Description
# ==========================================================================================
#
#   Constants for the Jacchia-Roberts 1971 Atmospheric Model.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

"""
    const _JR1971_CONSTANTS

Constants for the Jacchia-Roberts 1971  atmospheric model.
"""
const _JR1971_CONSTANTS = (;

    ########################################################################################
    #                                  General Constants
    ########################################################################################

    Rstar = 8.31432,     # ........ Rstar is the universal gas-constant [joules / (K . mol)]
    Av    = 6.022045e23, # ..................... Avogadro's constant (mks) [molecules / mol]
    Ra    = 6356.766,    # .......................................... Mean Earth radius [km]
    g₀    = 9.80665,     # ................................... Gravity at sea level [m / s²]

    ########################################################################################
    #                               Values at The Sea Level
    ########################################################################################

    M₀ = 28.960, # .............................. Mean molecular mass at sea level [g / mol]

    ########################################################################################
    #                               Values at Lower Boundary
    ########################################################################################

    T₁ = 183.0,    # .................................... Temperature at the lower bound [K]
    z₁ = 90.0,     # ...................................... Altitude of the lower bound [km]
    M₁ = 28.82678, # ................................. Mean molecular mass at `z₁` [g / mol]
    ρ₁ = 3.46e-9,  # ............................................. Density at `z₁` [g / cm³]

    ########################################################################################
    #                            Values at the Second Division
    ########################################################################################

    z₂ = 100.0,

    ########################################################################################
    #                            Values at the Inflection Point
    ########################################################################################

    zx = 125.0, # ................................... Altitude of the inflection point [km].

    ########################################################################################
    #                               Molecular Mass [g / mol]
    ########################################################################################

    Mi = (
        N₂ = 28.0134,
        O₂ = 31.9988,
        O  = 15.9994,
        Ar = 39.9480,
        He =  4.0026,
        H  = 1.00797
    ),

    ########################################################################################
    #                            Thermal Diffusion Coefficient
    ########################################################################################

    αi = (
        N₂ =  0,
        O₂ =  0,
        O  =  0,
        Ar =  0,
        He = -0.38,
        H  =  0
    ),

    ########################################################################################
    #                        Constituent density * M₀ / ρ_₁₀₀ / Av
    ########################################################################################

    μi = (
        N₂ = 0.78110,
        O₂ = 0.161778,
        O  = 0.095544,
        Ar = 0.0093432,
        He = 0.61471e-5,
        H  = 0
    ),

    ########################################################################################
    #                          Coefficients for Series Expansion
    ########################################################################################

    Aa = (
        -435093.363387,
          28275.5646391,
           -765.33466108,
             11.043387545,
             -0.08958790995,
              0.00038737586,
             -0.000000697444
    ),

    Ca = (
        -89284375.0,
          3542400.0,
           -52687.5,
              340.5,
               -0.8
    ),

    la = (
         0.1031445e+5,
         0.2341230e+1,
         0.1579202e-2,
        -0.1252487e-5,
         0.2462708e-9
    ),

    α = (
        3144902516.672729,
        -123774885.4832917,
           1816141.096520398,
            -11403.31079489267,
                24.36498612105595,
                 0.008957502869707995
    ),

    β = (
        -52864482.17910969,
           -16632.50847336828,
               -1.308252378125,
                0,
                0,
                0
    ),

    ζ = (
         0.1985549e-10,
        -0.1833490e-14,
         0.1711735e-17,
        -0.1021474e-20,
         0.3727894e-24,
        -0.7734110e-28,
         0.7026942e-32
    ),

    δij = (
        N₂ = (
             0.1093155e2,
             0.1186783e-2,  # (1/K)
            -0.1677341e-5,  # (1/K)²
             0.1420228e-8,  # (1/K)³
            -0.7139785e-12, # (1/K)⁴
             0.1969715e-15, # (1/K)⁵
            -0.2296182e-19  # (1/K)⁶
        ),

        O₂ = (
             0.9924237e1,
             0.1600311e-2,  # (1/K)
            -0.2274761e-5,  # (1/K)²
             0.1938454e-8,  # (1/K)³
            -0.9782183e-12, # (1/K)⁴
             0.2698450e-15, # (1/K)⁵
            -0.3131808e-19  # (1/K)⁶
        ),

        O = (
             0.1097083e2,
             0.6118742e-4,   # (1/K)
            -0.1165003e-6,   # (1/K)²
             0.9239354e-10,  # (1/K)³
            -0.3490739e-13,  # (1/K)⁴
             0.5116298e-17,  # (1/K)⁵
             0               # (1/K)⁶
        ),

        Ar = (
             0.8049405e1,
             0.2382822e-2,  # (1/K)
            -0.3391366e-5,  # (1/K)²
             0.2909714e-8,  # (1/K)³
            -0.1481702e-11, # (1/K)⁴
             0.4127600e-15, # (1/K)⁵
            -0.4837461e-19  # (1/K)⁶
        ),

        He = (
             0.7646886e1,
            -0.4383486e-3,  # (1/K)
             0.4694319e-6,  # (1/K)²
            -0.2894886e-9,  # (1/K)³
             0.9451989e-13, # (1/K)⁴
            -0.1270838e-16, # (1/K)⁵
             0              # (1/K)⁶
        ),
    )
)

"""
    const _JR1971_ROOT_GUESS

First guess to compute the roots of a polynomial to find the density below 125 km.
"""
const _JR1971_ROOT_GUESS = Complex{Float64}[166.10; 61.32; 9.91 + 1.31im; 9.91 - 1.31im]
